@@ -1,9 +1,9 @@
 from collections import OrderedDict
+
 from typing import Set, List, Optional, Union
 
-import gnupg
-
-from irrd.conf import get_setting, PASSWORD_HASH_DUMMY_VALUE
+from irrd.conf import PASSWORD_HASH_DUMMY_VALUE
+from irrd.utils.pgp import get_gpg_instance
 from .config import PASSWORD_HASHERS
 from .fields import (RPSLTextField, RPSLIPv4PrefixField, RPSLIPv4PrefixesField, RPSLIPv6PrefixField,
                      RPSLIPv6PrefixesField, RPSLIPv4AddressRangeField, RPSLASNumberField, RPSLASBlockField,
@@ -11,6 +11,11 @@ from .fields import (RPSLTextField, RPSLIPv4PrefixField, RPSLIPv4PrefixesField,
                      RPSLReferenceListField, RPSLAuthField, RPSLRouteSetMembersField, RPSLChangedField)
 from .parser import RPSLObject, UnknownRPSLObjectClassException
 
+RPSL_ROUTE_OBJECT_CLASS_FOR_IP_VERSION = {
+    4: 'route',
+    6: 'route6',
+}
+
 
 def rpsl_object_from_text(text, strict_validation=True, default_source: Optional[str]=None) -> RPSLObject:
     rpsl_object_class = text.split(':', maxsplit=1)[0].strip()
@@ -31,7 +36,7 @@ class RPSLAsBlock(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -47,7 +52,7 @@ class RPSLAsSet(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -71,7 +76,7 @@ class RPSLAutNum(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -90,7 +95,7 @@ class RPSLDomain(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, optional=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -106,7 +111,7 @@ class RPSLFilterSet(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -129,7 +134,7 @@ class RPSLInetRtr(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -147,7 +152,7 @@ class RPSLInet6Num(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -165,7 +170,7 @@ class RPSLInetnum(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -182,7 +187,7 @@ class RPSLKeyCert(RPSLObject):
         ('tech-c', RPSLReferenceField(lookup_key=True, optional=True, multiple=True, referring=['role', 'person'])),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -201,12 +206,12 @@ class RPSLKeyCert(RPSLObject):
         if not super().clean():
             return False  # pragma: no cover
 
-        gpg = gnupg.GPG(gnupghome=get_setting('auth.gnupg_keyring'))
+        gpg = get_gpg_instance()
         certif_data = '\n'.join(self.parsed_data.get('certif', [])).replace(',', '\n')
         result = gpg.import_keys(certif_data)
 
         if len(result.fingerprints) != 1:
-            msg = f'Unable to read public PGP key: key corrupt or multiple keys provided'
+            msg = 'Unable to read public PGP key: key corrupt or multiple keys provided'
             if result.results:
                 msg = f'{msg}: {result.results[0]["text"]}'
             self.messages.error(msg)
@@ -236,7 +241,7 @@ class RPSLKeyCert(RPSLObject):
     # which key signed a message, which can then be stored and compared to key-cert's later.
     # This method will probably be extracted to the update handler.
     def verify(self, message: str) -> bool:
-        gpg = gnupg.GPG(gnupghome=get_setting('auth.gnupg_keyring'))
+        gpg = get_gpg_instance()
         result = gpg.verify(message)
         return result.valid and result.key_status is None and \
             self.format_fingerprint(result.fingerprint) == self.parsed_data['fingerpr']
@@ -264,7 +269,7 @@ class RPSLMntner(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -342,7 +347,7 @@ class RPSLPeeringSet(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -358,7 +363,7 @@ class RPSLPerson(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -377,12 +382,14 @@ class RPSLRole(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
 
 class RPSLRoute(RPSLObject):
+    rpki_relevant = True
+    discarded_fields = ['rpki-ov-state']
     fields = OrderedDict([
         ('route', RPSLIPv4PrefixField(primary_key=True, lookup_key=True)),
         ('descr', RPSLTextField(multiple=True, optional=True)),
@@ -401,7 +408,7 @@ class RPSLRoute(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -418,12 +425,14 @@ class RPSLRouteSet(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
 
 class RPSLRoute6(RPSLObject):
+    rpki_relevant = True
+    discarded_fields = ['rpki-ov-state']
     fields = OrderedDict([
         ('route6', RPSLIPv6PrefixField(primary_key=True, lookup_key=True)),
         ('descr', RPSLTextField(multiple=True, optional=True)),
@@ -442,7 +451,7 @@ class RPSLRoute6(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -459,7 +468,7 @@ class RPSLRtrSet(RPSLObject):
         ('remarks', RPSLTextField(optional=True, multiple=True)),
         ('notify', RPSLEmailField(optional=True, multiple=True)),
         ('mnt-by', RPSLReferenceListField(lookup_key=True, multiple=True, referring=['mntner'])),
-        ('changed', RPSLChangedField(multiple=True)),
+        ('changed', RPSLChangedField(optional=True, multiple=True)),
         ('source', RPSLGenericNameField()),
     ])
 
@@ -484,6 +493,12 @@ OBJECT_CLASS_MAPPING = {
     'rtr-set': RPSLRtrSet,
 }
 
+RPKI_RELEVANT_OBJECT_CLASSES = [
+    rpsl_object.rpsl_object_class
+    for rpsl_object in OBJECT_CLASS_MAPPING.values()
+    if rpsl_object.rpki_relevant
+]
+
 
 def lookup_field_names() -> Set[str]:
     """Return all unique names of all lookup keys in all objects, plus 'origin'."""
