@@ -1,15 +1,22 @@
 import logging
 import re
+from collections import OrderedDict
+from typing import Optional, List, Set, Tuple, Any
+
+import ujson
 from IPy import IP
 from ordered_set import OrderedSet
-from typing import Optional, List, Set, Tuple
+from pytz import timezone
 
 from irrd import __version__
-from irrd.conf import get_setting
+from irrd.conf import get_setting, RPKI_IRR_PSEUDO_SOURCE
 from irrd.mirroring.nrtm_generator import NRTMGenerator, NRTMGeneratorException
-from irrd.rpsl.rpsl_objects import OBJECT_CLASS_MAPPING, lookup_field_names
-from irrd.storage.database_handler import DatabaseHandler
-from irrd.storage.preload import get_preloader
+from irrd.rpki.status import RPKIStatus
+from irrd.rpsl.rpsl_objects import (OBJECT_CLASS_MAPPING, lookup_field_names,
+                                    RPKI_RELEVANT_OBJECT_CLASSES)
+from irrd.scopefilter.status import ScopeFilterStatus
+from irrd.storage.database_handler import DatabaseHandler, is_serial_synchronised
+from irrd.storage.preload import Preloader
 from irrd.storage.queries import RPSLDatabaseQuery, DatabaseStatusQuery
 from irrd.utils.validators import parse_as_number, ValidationError
 from .query_response import WhoisQueryResponseType, WhoisQueryResponseMode, WhoisQueryResponse
@@ -42,17 +49,25 @@ class WhoisQueryParser:
     database_handler: DatabaseHandler
     _current_set_root_object_class: Optional[str]
 
-    def __init__(self, peer, peer_str: str) -> None:
-        self.all_valid_sources = list(get_setting('sources').keys())
+    def __init__(self, client_ip: str, client_str: str, preloader: Preloader,
+                 database_handler: DatabaseHandler) -> None:
+        self.all_valid_sources = list(get_setting('sources', {}).keys())
         self.sources_default = get_setting('sources_default')
         self.sources: List[str] = self.sources_default if self.sources_default else self.all_valid_sources
+        if get_setting('rpki.roa_source'):
+            self.all_valid_sources.append(RPKI_IRR_PSEUDO_SOURCE)
         self.object_classes: List[str] = []
         self.user_agent: Optional[str] = None
         self.multiple_command_mode = False
+        self.rpki_aware = bool(get_setting('rpki.roa_source'))
+        self.rpki_invalid_filter_enabled = self.rpki_aware
+        self.out_scope_filter_enabled = True
         self.timeout = 30
         self.key_fields_only = False
-        self.peer = peer
-        self.peer_str = peer_str
+        self.client_ip = client_ip
+        self.client_str = client_str
+        self.preloader = preloader
+        self.database_handler = database_handler
 
     def handle_query(self, query: str) -> WhoisQueryResponse:
         """
@@ -60,61 +75,67 @@ class WhoisQueryParser:
         Not thread safe - only one call must be made to this method at the same time.
         """
         # These flags are reset with every query.
-        self.database_handler = DatabaseHandler()
         self.key_fields_only = False
         self.object_classes = []
-        self.preloader = get_preloader()
 
         if query.startswith('!'):
             try:
                 return self.handle_irrd_command(query[1:])
             except WhoisQueryParserException as exc:
-                logger.info(f'{self.peer_str}: encountered parsing error while parsing query "{query}": {exc}')
+                logger.info(f'{self.client_str}: encountered parsing error while parsing query "{query}": {exc}')
                 return WhoisQueryResponse(
                     response_type=WhoisQueryResponseType.ERROR,
                     mode=WhoisQueryResponseMode.IRRD,
                     result=str(exc)
                 )
             except Exception as exc:
+                self.database_handler.refresh_connection()
+                # Do one retry after refreshing the DB connection
+                try:
+                    return self.handle_irrd_command(query[1:])
+                except Exception:
+                    pass
                 logger.error(f'An exception occurred while processing whois query "{query}": {exc}', exc_info=exc)
                 return WhoisQueryResponse(
                     response_type=WhoisQueryResponseType.ERROR,
                     mode=WhoisQueryResponseMode.IRRD,
                     result='An internal error occurred while processing this query.'
                 )
-            finally:
-                self.database_handler.close()
 
         try:
             return self.handle_ripe_command(query)
         except WhoisQueryParserException as exc:
-            logger.info(f'{self.peer_str}: encountered parsing error while parsing query "{query}": {exc}')
+            logger.info(f'{self.client_str}: encountered parsing error while parsing query "{query}": {exc}')
             return WhoisQueryResponse(
                 response_type=WhoisQueryResponseType.ERROR,
                 mode=WhoisQueryResponseMode.RIPE,
                 result=str(exc)
             )
         except Exception as exc:
+            self.database_handler.refresh_connection()
+            # Do one retry after refreshing the DB connection
+            try:
+                return self.handle_ripe_command(query)
+            except Exception:
+                pass
             logger.error(f'An exception occurred while processing whois query "{query}": {exc}', exc_info=exc)
             return WhoisQueryResponse(
                 response_type=WhoisQueryResponseType.ERROR,
                 mode=WhoisQueryResponseMode.RIPE,
                 result='An internal error occurred while processing this query.'
             )
-        finally:
-            self.database_handler.close()
 
     def handle_irrd_command(self, full_command: str) -> WhoisQueryResponse:
         """Handle an IRRD-style query. full_command should not include the first exclamation mark. """
         if not full_command:
-            raise WhoisQueryParserException(f'Missing IRRD command')
-        command = full_command[0].upper()
+            raise WhoisQueryParserException('Missing IRRD command')
+        command = full_command[0]
         parameter = full_command[1:]
         response_type = WhoisQueryResponseType.SUCCESS
         result = None
 
         # A is not tested here because it is already handled in handle_irrd_routes_for_as_set
-        queries_with_parameter = list('TG6IJMNORS')
+        queries_with_parameter = list('tg6ijmnors')
         if command in queries_with_parameter and not parameter:
             raise WhoisQueryParserException(f'Missing parameter for {command} query')
 
@@ -122,11 +143,19 @@ class WhoisQueryParser:
             self.multiple_command_mode = True
             result = None
             response_type = WhoisQueryResponseType.NO_RESPONSE
-        elif command == 'V':
+        elif full_command.upper() == 'FNO-RPKI-FILTER':
+            self.rpki_invalid_filter_enabled = False
+            result = 'Filtering out RPKI invalids is disabled for !r and RIPE style ' \
+                     'queries for the rest of this connection.'
+        elif full_command.upper() == 'FNO-SCOPE-FILTER':
+            self.out_scope_filter_enabled = False
+            result = 'Filtering out out-of-scope objects is disabled for !r and RIPE style ' \
+                     'queries for the rest of this connection.'
+        elif command == 'v':
             result = self.handle_irrd_version()
-        elif command == 'T':
+        elif command == 't':
             self.handle_irrd_timeout_update(parameter)
-        elif command == 'G':
+        elif command == 'g':
             result = self.handle_irrd_routes_for_origin_v4(parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
@@ -134,31 +163,33 @@ class WhoisQueryParser:
             result = self.handle_irrd_routes_for_origin_v6(parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
-        elif command == 'A':
+        elif command == 'a':
             result = self.handle_irrd_routes_for_as_set(parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
-        elif command == 'I':
+        elif command == 'i':
             result = self.handle_irrd_set_members(parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
-        elif command == 'J':
+        elif command == 'j':
             result = self.handle_irrd_database_serial_range(parameter)
-        elif command == 'M':
+        elif command == 'J':
+            result = self.handle_irrd_database_status(parameter)
+        elif command == 'm':
             result = self.handle_irrd_exact_key(parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
-        elif command == 'N':
+        elif command == 'n':
             self.handle_user_agent(parameter)
-        elif command == 'O':
+        elif command == 'o':
             result = self.handle_inverse_attr_search('mnt-by', parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
-        elif command == 'R':
+        elif command == 'r':
             result = self.handle_irrd_route_search(parameter)
             if not result:
                 response_type = WhoisQueryResponseType.KEY_NOT_FOUND
-        elif command == 'S':
+        elif command == 's':
             result = self.handle_irrd_sources_list(parameter)
         else:
             raise WhoisQueryParserException(f'Unrecognised command: {command}')
@@ -215,7 +246,7 @@ class WhoisQueryParser:
             ip_version = 6
 
         if not set_name:
-            raise WhoisQueryParserException(f'Missing required set name for A query')
+            raise WhoisQueryParserException('Missing required set name for A query')
 
         self._current_set_root_object_class = 'as-set'
         members = self._recursive_set_resolve({set_name})
@@ -398,12 +429,11 @@ class WhoisQueryParser:
         for query_result in query_results:
             source = query_result['source'].upper()
             keep_journal = 'Y' if get_setting(f'sources.{source}.keep_journal') else 'N'
-            serial_oldest = query_result['serial_oldest_seen']
-            serial_newest = query_result['serial_newest_seen']
+            serial_newest = query_result['serial_newest_mirror']
             fields = [
                 source,
                 keep_journal,
-                f'{serial_oldest}-{serial_newest}' if serial_oldest and serial_newest else '-',
+                f'0-{serial_newest}' if serial_newest else '-',
             ]
             if query_result['serial_last_export']:
                 fields.append(str(query_result['serial_last_export']))
@@ -413,6 +443,37 @@ class WhoisQueryParser:
             result_txt += f'{invalid_source.upper()}:X:Database unknown\n'
         return result_txt.strip()
 
+    def handle_irrd_database_status(self, parameter: str) -> str:
+        """!J query - database status"""
+        if parameter == '-*':
+            sources = self.sources_default if self.sources_default else self.all_valid_sources
+        else:
+            sources = [s.upper() for s in parameter.split(',')]
+        invalid_sources = [s for s in sources if s not in self.all_valid_sources]
+        query = DatabaseStatusQuery().sources(sources)
+        query_results = self.database_handler.execute_query(query)
+
+        results: OrderedDict[str, OrderedDict[str, Any]] = OrderedDict()
+        for query_result in query_results:
+            source = query_result['source'].upper()
+            results[source] = OrderedDict()
+            results[source]['authoritative'] = get_setting(f'sources.{source}.authoritative', False)
+            object_class_filter = get_setting(f'sources.{source}.object_class_filter')
+            results[source]['object_class_filter'] = list(object_class_filter) if object_class_filter else None
+            results[source]['rpki_rov_filter'] = bool(get_setting('rpki.roa_source') and not get_setting(f'sources.{source}.rpki_excluded'))
+            results[source]['scopefilter_enabled'] = not get_setting(f'sources.{source}.scopefilter_excluded')
+            results[source]['local_journal_kept'] = get_setting(f'sources.{source}.keep_journal', False)
+            results[source]['serial_oldest_journal'] = query_result['serial_oldest_journal']
+            results[source]['serial_newest_journal'] = query_result['serial_newest_journal']
+            results[source]['serial_last_export'] = query_result['serial_last_export']
+            results[source]['serial_newest_mirror'] = query_result['serial_newest_mirror']
+            results[source]['last_update'] = query_result['updated'].astimezone(timezone('UTC')).isoformat()
+            results[source]['synchronised_serials'] = is_serial_synchronised(self.database_handler, source)
+
+        for invalid_source in invalid_sources:
+            results[invalid_source.upper()] = OrderedDict({'error': 'Unknown source'})
+        return ujson.dumps(results, indent=4)
+
     def handle_irrd_exact_key(self, parameter: str):
         """!m query - exact object key lookup, e.g. !maut-num,AS65537"""
         try:
@@ -593,13 +654,13 @@ class WhoisQueryParser:
     def handle_user_agent(self, user_agent: str):
         """-V/!n parameter/query - set a user agent for the client"""
         self.user_agent = user_agent
-        logger.info(f'{self.peer_str}: user agent set to: {user_agent}')
+        logger.info(f'{self.client_str}: user agent set to: {user_agent}')
 
     def handle_nrtm_request(self, param):
         try:
             source, version, serial_range = param.split(':')
         except ValueError:
-            raise WhoisQueryParserException(f'Invalid parameter: must contain three elements')
+            raise WhoisQueryParserException('Invalid parameter: must contain three elements')
 
         try:
             serial_start, serial_end = serial_range.split('-')
@@ -618,8 +679,8 @@ class WhoisQueryParser:
         if source not in self.all_valid_sources:
             raise WhoisQueryParserException(f'Unknown source: {source}')
 
-        if not is_client_permitted(self.peer, f'sources.{source}.nrtm_access_list'):
-            raise WhoisQueryParserException(f'Access denied')
+        if not is_client_permitted(self.client_ip, f'sources.{source}.nrtm_access_list'):
+            raise WhoisQueryParserException('Access denied')
 
         try:
             return NRTMGenerator().generate(source, version, serial_start, serial_end, self.database_handler)
@@ -646,11 +707,15 @@ class WhoisQueryParser:
         if self.sources and self.sources != self.all_valid_sources:
             query.sources(self.sources)
         else:
-            default = get_setting('sources_default')
+            default = list(get_setting('sources_default', []))
             if default:
                 query.sources(list(default))
         if self.object_classes:
             query.object_classes(self.object_classes)
+        if self.rpki_invalid_filter_enabled:
+            query.rpki_status([RPKIStatus.not_found, RPKIStatus.valid])
+        if self.out_scope_filter_enabled:
+            query.scopefilter_status([ScopeFilterStatus.in_scope])
         return query
 
     def _execute_query_flatten_output(self, query: RPSLDatabaseQuery) -> str:
@@ -664,11 +729,21 @@ class WhoisQueryParser:
         else:
             result = ''
             for obj in query_response:
-                result += obj['object_text'] + '\n'
+                result += obj['object_text']
+                if (
+                        self.rpki_aware and
+                        obj['source'] != RPKI_IRR_PSEUDO_SOURCE and
+                        obj['object_class'] in RPKI_RELEVANT_OBJECT_CLASSES
+                ):
+                    comment = ''
+                    if obj['rpki_status'] == RPKIStatus.not_found:
+                        comment = ' # No ROAs found, or RPKI validation not enabled for source'
+                    result += f'rpki-ov-state:  {obj["rpki_status"].name}{comment}\n'
+                result += '\n'
         return result.strip('\n\r')
 
     def _filter_key_fields(self, query_response) -> str:
-        results = OrderedSet()
+        results: OrderedSet[str] = OrderedSet()
         for obj in query_response:
             result = ''
             rpsl_object_class = OBJECT_CLASS_MAPPING[obj['object_class']]
